from ..utils import fetch, dictionary_to_string
from ..model import ScholarProfile, ThumbnailProfile
from dataclasses import make_dataclass, field
from random import choice
from typing import (List, Dict, Tuple, Any, Union)
from types import FunctionType


class Engine(object):
    def __init__(self,
                 seconds: int = 5,
                 http: Dict[str, str] = None,
                 https: Dict[str, str] = None):
        self.seconds: int = seconds
        self.http: List[str] = http
        self.https: List[str] = https

    @property
    def proxies(self) -> Dict[str, str]:
        '''Random selector of proxies for request.'''
        if self.http is None and self.https is None:
            return None
        elif self.http and self.https:
            return {'http': choice(self.http), 'https': choice(self.https)}
        elif self.http:
            return {'http': choice(self.http)}
        elif self.https:
            return {'https': choice(self.https)}

    def _request(self, url: str, **kwargs: Dict[str, str]) -> Any:
        '''
        GET request and parsing the response with a callback.

        Arguments:
            - **kwargs:
                - callback: A parsing for function for the resulting response
                - params: A dictionary with addiitional parameter to be sent along with GET request
        '''
        callback: FunctionType = kwargs.get('callback', None)
        params: Dict[str, str] = kwargs.get('params', None)
        meta: Dict[str, Any] = kwargs.get('meta', None)

        # Sending GET request
        # . Response is converted into BeautifulSoup object within fetch
        response: 'bs4.BeautifulSoup' = fetch(url,
                                              params=params,
                                              seconds=self.seconds,
                                              proxies=self.proxies)
        # Calling a parser on BeautifulSoup response
        if callback:
            return callback(response, meta=meta)
        return response

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({dictionary_to_string(vars(self))})'

    @staticmethod
    def alter_dataclass(obj: 'dataclass', *,
                        attributes: Union[Dict[str, Any], List[Tuple[Any]]]
                        ) -> 'dataclass':
        '''
        Extends a given dataclass to include additional attributes.

        Arguments:
            - obj: A dataclass object
            - attributes: The attributes to be added to obj
        '''
        try:
            if isinstance(attributes, list):
                attributes = dict(attributes)
        except Exception as e:
            raise type(e)(
                'Attributes must be a dictionary or a list of tuples.')
        else:
            obj.__class__ = make_dataclass(obj.__class__.__name__,
                                           fields=[
                                               (k, type(v),
                                                field(default_factory=type(v),
                                                      repr=False))
                                               for k, v in attributes.items()
                                           ],
                                           bases=(obj.__class__, ))
            vars(obj).update(attributes)
            return obj
