import re

from .engine import Engine
from ..parsers import thmbnl_profile_loader, scholar_profile_loader, paper_loader
from ..model import ThumbnailProfile, ScholarProfile, Paper
from bs4 import BeautifulSoup
from itertools import chain
from typing import (List, Dict, Tuple, Any, Union)
from urllib.parse import urlencode, urlparse, quote_plus


class GoogleScholar(Engine):
    GENERAL_URL: str = 'https://scholar.google.ca'
    AUTHORS_URL: str = 'https://scholar.google.ca/citations?'

    # TODO: **fields
    def search(self, query: str) -> List[ThumbnailProfile]:
        '''
        Recovers first 10 search results of a given author search.

        Arguments:
            - query: A string of values to be search on Google Scholar authors
                - Can specify 'label: \{query\}' for filtering by tags
            - **fields: Additional fields to include in ThumbnailProfile

        Example:
            >>> scholar = GoogleScholar()

            >>> scholar.search('John Smith')
            [ThumbnailProfile(name='', affiliation='', num_citation=0, email_domain='', tags=[])]
            >>> scholar.search('John Smith label:Ecology')
            [ThumbnailProfile(name='', affiliation='', num_citation=0, email_domain='', tags=['Ecology'])]
        '''
        thmbnl_profile: ThumbnailProfile = self._request(
            url=self.AUTHORS_URL,
            callback=thmbnl_profile_loader,
            params=self._authors_query(query),
        )
        return thmbnl_profile

    def search_by(self, papers=False,
                  **kwargs: Dict[str, Any]) -> ScholarProfile:
        '''
        Searchers profile by unique identifier or URL.

        Arguments:
            - papers: Boolean flag determining if papers are scraped
            - **kwargs:
                - _id: A unique profile identifier
                - url: A profile's URL
                - Any other arguments will be added as new attributes to the resulting ScholarProfile instance

        Exceptions:
            - ValueError: Neither a unique identifier or URL are specified

        Examples:
            >>> from woden.engine import GoogleScholar
            >>> scholar = GoogleScholar()
            >>> scholar.search(_id='XC-SDFSDSF')
            ScholarProfile()
            >>> scholar.search(url='https://scholar.google.com/citations?...')
            ScholarProfile()
            >>> sub_scholar = scholar.search(_id='XC-SDFSDSF', fields={'field1': 1})
            >>> print(sub_scholar.field1)
            1
            >>> sub_scholar = scholar.saerch(url='https://scholar.google.com/citations?...', fields={'fields1': 1})
            >>> print(sub_scholar.field1)
            1
        '''
        if not any(key in kwargs for key in ('_id', 'url')):
            raise ValueError(
                'Must pass a unique identifer (_id) or a URL (url).')

        if '_id' in kwargs:
            url, scholar_profile = self._search_by_id(kwargs['_id'])
        elif 'url' in kwargs:
            url = kwargs['url']
            scholar_profile: ScholarProfile = self._search_by_url(url)

        if papers:
            scholar_profile.papers: List[Paper] = list(
                chain(*self._collect_papers(url)))

        if 'fields' in kwargs:
            return self.alter_dataclass(scholar_profile,
                                        attributes=kwargs['fields'])
        return scholar_profile

    def _search_by_id(self, _id: str) -> ScholarProfile:
        '''
        Searches profile by a profile's unique identifier.

        Arguments:
            - _id: The unique identified of a profile
        '''
        # The profile unique identifier to request a response from, or None
        query: Dict[str, str] = urlencode(self._id_query(_id))
        url: str = self.AUTHORS_URL + query
        return url, self._request(url,
                                  callback=scholar_profile_loader,
                                  meta={'ref': query})

    def _search_by_url(self, url: str) -> ScholarProfile:
        '''
        Searches profile by a a direct URL.

        Arguments:
            - url: URL of a profile 
        '''
        return self._request(url,
                             callback=scholar_profile_loader,
                             meta={'ref': urlparse(url).query})

    def _collect_papers(self, url: str) -> List[Paper]:
        '''
        Collects papers on a profile.

        Arguments:
            - url: Profile's URL
        '''
        if url is None:
            raise ValueError('URL cannot be NoneType.')
        cstart: int = 0
        response = BeautifulSoup(u'', 'html.parser')
        while not response.find('td', {'class': 'gsc_a_e'}):
            response = self._request(url, params=self._paper_query(cstart))
            yield paper_loader(response)
            cstart += 100

    @staticmethod
    def _paper_query(cstart: int) -> Dict[str, str]:
        '''
        Creates a dictionary of paramerts for a GET request.

        Arguments:
            - cstart: The paper index to begin listing from
        '''
        return {'hl': 'en', 'cstart': cstart, 'pagesize': '100'}

    @staticmethod
    def _authors_query(mauthors: str) -> Dict[str, str]:
        '''
        Creates a dictionary of parameters for a GET request.

        Arguments:
            - mauthors: The string to be searched

        Notes:
            - 'label: \{query\}' can be used to specify tags in search
        '''
        return {
            'view_op': 'search_authors',
            'hl': 'en',
            'mauthors': quote_plus(mauthors)
        }

    @staticmethod
    def _id_query(_id: str) -> Dict[str, str]:
        '''
        Creates a dictionary of parameters for a GET request.

        Arguments:
            - _id: The unique identifier of a Google Scholar profile
        '''
        return {'hl': 'en', 'user': _id}
