import re

from bs4 import BeautifulSoup
from ..model import ThumbnailProfile
from ..utils import ignore, lower_txt, select, select_href
from typing import List, Generator, Tuple, Dict, Union


DOMAIN = re.compile('verified email at (.*)')

def thmbnl_profile_loader(soup: BeautifulSoup,
                          meta=None) -> List[ThumbnailProfile]:
    '''Parses a BeautifulSoup object, returns a list of ThumbnailProfile objects'''
    names: List[str] = select(soup, '.gs_ai_name')
    affiliations: List[str] = select(soup, '.gs_ai_aff')
    tags: List[str] = list(_tags(select(soup, 'gs_ai_int')))
    num_citations: List[str] = list(_citations( select(soup, '.gs_ai_cby')))
    email_domains: List[str] = list(_email_domain_matcher(select(soup, '.gs_ai_eml')))
    ref: List[str] = select_href(soup, 'gs_ai_name a')

    profile_dicts: Generator[Tuple[str], None, None] = _profile_keyworder(
        zip(names, affiliations, email_domains, tags, num_citations, ref))

    return [ThumbnailProfile(**fields) for fields in profile_dicts]

def _hrefs(a_tags):
    return [a['href'].replace('/citations?', '') for a in a_tags]

def _email_domain_matcher(email_domains: str
                          ) -> Generator[Union[str, None], None, None]:
    '''Returns a string from a Match object or None'''
    for match in map(DOMAIN.match, email_domains):
        if match is not None:
            yield match.group(1)
        else:
            yield match


def _profile_keyworder(profile_tuples: Generator[Tuple[str], None, None]
                       ) -> Generator[Dict[str, str], None, None]:
    '''Creates dictionary representations of ThumbnailProfile.'''
    fields: Tuple[str] = ('name', 'affiliation', 'email_domain', 'tags',
                          'num_citations')
    for profile_tuple in profile_tuples:
        yield dict(zip(fields, profile_tuple))


def _tags(tag_soup: BeautifulSoup) -> List[str]:
    '''Return lower case tags.'''
    for tags in map(lambda x: x.select('.gs_ai_one_int'), tag_soup):
        yield [lower_txt(tag) for tag in tags]


@ignore(TypeError)
def _citations(citation_soup: BeautifulSoup) -> str:
    '''Returns number of citations for a profile.'''
    for cit in citation_soup:
        yield cit.get_text().replace('Cited by ', '')
